/*****************************************************************************
 *
 *                        CalC (Calcium Calculator)
 *                             Victor Matveev
 *
 *                               field.cpp
 *
 *  Definitions of the main FieldObj concentration field class (derived from
 *  the VectorObj class), the daughter BufferObj buffer concentration field 
 *  class, and the compound buffer array class BufferArray.
 *
 *  FieldObj methods include all low-level ADI difference schemes routines.
 *  Higher-level numeric engine routines are defined in "calc.cpp"       
 *       
 *  FieldObj and BufferArray variables are member of the main SimulationObj
 *  object, defined in simulation.h / simulation.cpp
 *
 ******************************************************************************/

#include "stdafx.h"
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include "vector.h"
#include <stdarg.h>
#include "syntax.h"
#include "box.h"
#include "grid.h"
#include "field.h"

extern void setTortuosity(TokenString &, long, double *);                                     

const double pi = atan(1.0) * 4;

//****************************************************************************
//                      INITIALIZING STATIC MEMBERS:
//****************************************************************************

class RegionObj   *FieldObj::Region = 0;
class GridObj     *FieldObj::Grid = 0;
class BCarrayObj  *FieldObj::BCarray = 0;

int    FieldObj::xsize = 0,   FieldObj::ysize = 0,    FieldObj::zsize = 0;
long   FieldObj::Size = 0,    FieldObj::xysize = 0;

double *FieldObj::xgrid = 0,  *FieldObj::ygrid = 0,   *FieldObj::zgrid = 0;
double *FieldObj::xcoord = 0, *FieldObj::ycoord = 0,  *FieldObj::zcoord = 0;

double *FieldObj::dxplus = 0, *FieldObj::dxminus = 0, *FieldObj::dxinv = 0, *FieldObj::dx2inv = 0;
double *FieldObj::dyplus = 0, *FieldObj::dyminus = 0, *FieldObj::dyinv = 0, *FieldObj::dy2inv = 0;
double *FieldObj::dzplus = 0, *FieldObj::dzminus = 0, *FieldObj::dzinv = 0, *FieldObj::dz2inv = 0;
  
double *FieldObj::bc_deriv = 0, *FieldObj::bc_coef = 0, *FieldObj::bc_lin = 0;
char   **FieldObj::bc_id = 0;
int    FieldObj::bc_type_num = 0;

double *FieldObj::right = 0, *FieldObj::diag = 0, *FieldObj::sup = 0, *FieldObj::sub = 0;

nabla  FieldObj::nablaX = 0, FieldObj::nablaY = 0, FieldObj::nablaZ = 0;

//*************************************************************************
//*                T R I D I A G O N A L   S Y S T E M
//*************************************************************************

void FieldObj::setStaticData(RegionObj &RO, GridObj &GO, BCarrayObj &BCO) {

  Region = &RO;

  Grid = &GO;
  Size = GO.Size;    xysize = GO.xysize; 
  xsize = GO.xsize;  ysize = GO.ysize; zsize = GO.zsize;

  xgrid = GO.xgrid;   ygrid = GO.ygrid;   zgrid = GO.zgrid; 
  xcoord = GO.xcoord; ycoord = GO.ycoord; zcoord = GO.zcoord; 

  dxplus = GO.dxplus;   dyplus = GO.dyplus;   dzplus = GO.dzplus; 
  dxminus = GO.dxminus; dyminus = GO.dyminus; dzminus = GO.dzminus;
  dxinv = GO.dxinv;     dyinv = GO.dyinv;     dzinv = GO.dzinv;  
  dx2inv = GO.dx2inv;   dy2inv = GO.dy2inv;   dz2inv = GO.dz2inv;  

  BCarray = &BCO;
  bc_id = BCO.bc_id;        bc_type_num = BCO.bc_type_num;
  bc_deriv = BCO.bc_deriv;  bc_coef=BCO.bc_coef;  bc_lin = BCO.bc_lin;

  int max = ( xsize > ysize ? xsize : ysize );
  if (zsize > max) max = zsize;
  init_tridiag(max + 2);
} 


void FieldObj::init_tridiag(int n)
{
diag = new double[n];  right = new double[n];
sup = new double[n];   sub = new double[n];
}

void FieldObj::kill_tridiag()
{
  if (diag) {
    delete [] diag; delete [] right;
    delete [] sup;  delete [] sub;
  }
}

//*******************   S O L V E R :  **************************

void FieldObj::tridiag(int n)
{
int    i;
register double sb, sp, dg, rt;

sp = ( sup[1] /= (dg = diag[1]) );
rt = ( right[1] /= dg );

for (i=2; i<=n; i++)
  {
  sb = sub[i];
  rt = right[i] = (right[i] - sb * rt) / (dg = diag[i] - sb * sp);
  sp = (sup[i] /= dg);
  }

for (i=n-1; i>=1; i--) right[i] -= sup[i] * right[i+1];
}

//****************************************************************************
//*          C L A S S   F I E L D :   I M P L E M E N T A T I O N
//****************************************************************************

void FieldObj::allocate(const char *id)
{
ID = StrCpy(id);

bccond = new int[Region -> get_surface_num()];

  ptype = new long[size];

  source_x0    = new int[source_num];
  source_nx    = new int[source_num];
  source_y0    = new int[source_num];
  source_ny    = new int[source_num];
  source_z0    = new int[source_num];
  source_nz    = new int[source_num];
  weights      = new double [source_num];
  source_xsum  = new double *[source_num];
  source_ysum  = new double *[source_num];
  source_zsum  = new double *[source_num];

}


//****************************************************************************


FieldObj::FieldObj(TokenString &params, const char *id) : VectorObj(Size)
{
  int i;

  source_num = params.token_count(id, "source");
  //params.checkName(id, "concentration field");
  allocate(id);

  Time = 0.0;  current = 0; 
  *this = total = bgr = 0.0;

  params.get2_param(id, "D", &D, " undefined coefficient of diffusion");   
  if (VERBOSE)
    fprintf(stdlog, "\n### Concentration field %s: diffusion coefficient = %g um^2 / ms\n", ID, D);

  tortDefined = false;
  long tortIndex = 0;                             // this blocks sets tortuosity
  Diff = 0;
  char *idtort = new char[ strlen(id) + 12 ];
  strcpy(idtort, id); strcat(idtort,".tortuosity");
  params.token_count(idtort, &tortIndex);
  if ( tortIndex ) {
    tortDefined = true;
    if ( params.equal( ++tortIndex, "=") ) ++tortIndex;
    if (VERBOSE) fprintf(stdlog,"    Tortuosity function specified: \n");
    Diff = new double[ Size ];
    setTortuosity( params, tortIndex, Diff ); 
  }
  delete [] idtort;

  if (params.token_count(id, "total")) {
    total = params.get2_param(id, "total");
    *this = total;
    if (equal(id, "Ca") ) throw
      makeMessage("For Ca, \"bgr\" (background) concentration has to be defined instead of \"total\"");
    if (VERBOSE) fprintf(stdlog, "    Total [%s] = %g uM \n", ID, total);
  }
  else if ( !equal(id, "Ca") ) 
      throw makeMessage("Total concentration not defined for buffer \"%s\"", id);

  if (params.token_count(id, "bgr")) {
    bgr = params.get2_param(id, "bgr");
    *this = bgr;
    if (VERBOSE)  fprintf(stdlog, "    Background [%s] = %g uM \n", ID, bgr);
  }

  long pos;
  if (params.token_count(id, "import", &pos))
    { 
    char *filename = params.get_string(pos + 1);
    if (VERBOSE) fprintf(stdlog, "  # Importing %s from file %s\n", ID, filename);
    import(filename);
    delete filename;
    }

 if (source_num)
  {
  double x, y, z, sigmax, sigmay, sigmaz, weight;
  if (VERBOSE) fprintf(stdlog, "    %s has %d sources:\n", ID, source_num);
  for (i = 1; i <= source_num; i++)
  try {
    x = y = z = 0;
    sigmax = 0.0; sigmay = sigmaz = -1;
    weight = 1.0;

    switch (GEOMETRY & DIMENSION_MASK) {
    case 1:
       params.trail_pars(id, "source", i, 'd', &x, 'd', &sigmax, 'd', &weight);
       break;
    case 2: 
       params.trail_pars(id, "source", i, 'd', &x, 'd', &y, 'd', &sigmax, 'd', &sigmay, 'd', &weight);
       break;
    case 3:
       params.trail_pars(id, "source", i, 'd', &x, 'd', &y, 'd', &z, 'd', &sigmax, 
                          'd', &sigmay,  'd', &sigmaz, 'd', &weight);
       break;
    }

    if (sigmay < 0) sigmay = sigmax;  //  by default widths in all directions are equal
    if (sigmaz < 0) sigmaz = sigmay;  //

    weights[i-1] = weight;
    set_source(i-1, x, y, z, sigmax, sigmay, sigmaz); 
    } 
    catch (char *str) { 
      params.errorMessage( params.token_index(id, "source", i) + 1, str, "Cannot initialize this source"); 
    }
  //    catch (int m) { 
  //    params.errorMessage( params.token_index(id, "source", i) + 1, 0, "Cannot initialize this source"); 
  //  }
   }

  int defs  = params.token_count(id, "bc");
  int boxes = Region->get_box_num();

  char bc1[20], bc2[20], bc3[20], bc4[20], bc5[20], bc6[20];

  for (i = 1; i<= boxes; i++)
    try { 
    if ( i > defs )   // no b.c. specification for box i:
      {
      if (VERBOSE) fprintf(stdlog,"    Box #%d: not specified; assuming all boundaries are Neumann\n", i);
      set_bc(i-1, "Neumann");     
      }
    else if ( params.equal( params.token_index(id, "bc", i) + 1, "all") ) 
      {
      params.trail_pars(id, "bc", i, 0,0, 's', bc1, 'E');
      if (VERBOSE) fprintf(stdlog,"    Box #%d: all boundaries are %s\n", i, bc1);
      set_bc(i-1, bc1);
      }
    else if ((GEOMETRY & DIMENSION_MASK) == 1)
      {
      params.trail_pars(id, "bc", i, 's', bc1, 's', bc2, 'E');
      if (VERBOSE) 
        fprintf(stdlog,"    Box #%d: boundary cond'ns: xmin:%s xmax:%s\n", i, bc1, bc2);
      set_bc(i-1, bc1, bc2, bc2, bc2, bc2, bc2);
      }
    else if ((GEOMETRY & DIMENSION_MASK) == 2)
      {
      params.trail_pars(id, "bc", i, 's', bc1, 's', bc2, 's', bc3, 's', bc4,  'E');
      if (VERBOSE) 
        fprintf(stdlog,"    Box #%d: boundary cond'ns: xmin:%s xmax:%s ymin:%s ymax:%s\n", i, bc1, bc2, bc3, bc4);
      set_bc(i-1, bc1, bc2, bc3, bc4, bc4, bc4);
      }
    else  // DIMENSIONALITY == 3
      {
      params.trail_pars(id, "bc", i,
          's', bc1, 's', bc2, 's', bc3, 's', bc4, 's', bc5, 's', bc6, 'E');
      if (VERBOSE) 
        fprintf(stdlog,"    Box #%d: boundary cond'ns: xmin:%s xmax:%s ymin:%s ymax:%s zmin:%s zmax:%s\n", 
                i, bc1, bc2, bc3, bc4, bc5, bc6);
      set_bc(i-1, bc1, bc2, bc3, bc4, bc5, bc6);
      }
    }
  catch(char *str) { params.errorMessage( params.token_index(id, "bc", i) + 1, str); }

  init_boundary();
  // initialAve = bgr;
}

//***************************************************************************

FieldObj::FieldObj(const FieldObj &f) : VectorObj(f)
{
  int nx, ny, nz;
  source_num = f.source_num;
  allocate( f.ID );
 
  current = f.current;

  if (source_num)
    {
    for(int i = 0; i < source_num; i++)  
      {
      weights[i] = f.weights[i];
      source_x0[i] = f.source_x0[i];
      nx = source_nx[i] = f.source_nx[i];
      source_y0[i] = f.source_y0[i];
      ny = source_ny[i] = f.source_ny[i];
      source_z0[i] = f.source_z0[i];
      nz = source_nz[i] = f.source_nz[i];
      
      source_xsum[i] = new double[nx];
      source_ysum[i] = new double[ny];
      source_zsum[i] = new double[nz];

      int ii;
      for (ii = 0; ii < nx; ii++) source_xsum[i][ii] = f.source_xsum[i][ii];
      for (ii = 0; ii < ny; ii++) source_ysum[i][ii] = f.source_ysum[i][ii];
      for (ii = 0; ii < nz; ii++) source_zsum[i][ii] = f.source_zsum[i][ii];
      }
    }

  for(long i=0; i< size; i++) ptype[i] = f.ptype[i];
  for(int n=0; n< Region->get_surface_num(); n++)  bccond[n] = f.bccond[n];

  D = f.D;
  Time = f.Time;
  total = f.total;
  bgr = f.bgr;
  Diff = f.Diff;
  // initialAve = f.initialAve;
}

//***************************************************************************

FieldObj &FieldObj::operator=(const FieldObj &f)
{
 *((VectorObj *)this) = f;
 Time = f.Time;
 return *this;
}

//***************************************************************************

FieldObj &FieldObj::operator=(const VectorObj &v)
{
*((VectorObj *)this) = v;

//for (long i = 0; i < size; ++i) elem[i] = v[i]; 

return *this;
}

//***************************************************************************

FieldObj &FieldObj::operator=(const double val)
{
*((VectorObj *)this) = val;

//for (long i = 0; i < size; ++i) elem[i] = val; 
return *this;
}

//************************************************************************

FieldObj &FieldObj::operator=(double (*f0)(double, double, double) )
{
long   i = -1;

for (int iz = 0; iz < zsize; iz++)
  for (int iy = 0; iy < ysize; iy++)
    for (int ix = 0; ix < xsize; ix++)
      if ( ptype[++i] & VARY_MASK )
         elem[i] = (*f0)(xcoord[ix], ycoord[iy], zcoord[iz]);
 
return *this;
}


//************************************************************************************
//                   F I E L D   D E S T R U C T O R
//************************************************************************************

  FieldObj::~FieldObj()
    {
    delete [] ptype;
    delete [] bccond; 
 
    for (int i = 0; i < source_num; i++) 
      {
      delete [] source_xsum[i]; delete [] source_ysum[i]; delete [] source_zsum[i];
      }

    delete [] weights;
    delete [] source_x0;   delete [] source_nx;
    delete [] source_y0;   delete [] source_ny;
    delete [] source_z0;   delete [] source_nz; 

    delete [] ID;
  }


//************************************************************************************
//                   F I E L D   I N I T I A L I Z A T I O N
//************************************************************************************


void  FieldObj::set_bc(int boxid, int bcx0, int bcx1, int bcy0, int bcy1, int bcz0, int bcz1)
{
  if (boxid >= Region->get_box_num())  
    throw makeMessage("Cannot set boundary condition: box ID=%d > total number of volumes = %d", 
                      boxid, Region->get_box_num());

  if ( (bcx0 > bc_type_num) || (bcx1 > bc_type_num) || (bcy0 > bc_type_num) ||
       (bcy1 > bc_type_num) || (bcz0 > bc_type_num) || (bcz1 > bc_type_num) )  
    throw makeMessage("Bad boundary condition ID for geometry element #%d", boxid);
  
  bccond[boxid*6]   = bcx0;
  bccond[boxid*6+1] = bcx1;
  bccond[boxid*6+2] = bcy0;
  bccond[boxid*6+3] = bcy1;
  bccond[boxid*6+4] = bcz0;
  bccond[boxid*6+5] = bcz1;

}

//******************************************

void  FieldObj::set_bc(int boxid, const char *bcx0, const char *bcx1, const char *bcy0,
                                  const char *bcy1, const char *bcz0, const char *bcz1)
{
  set_bc( boxid, BCarray->bcidtoint(bcx0), BCarray->bcidtoint(bcx1), 
                 BCarray->bcidtoint(bcy0), BCarray->bcidtoint(bcy1), 
	         BCarray->bcidtoint(bcz0), BCarray->bcidtoint(bcz1) );
}

//******************************************

void  FieldObj::set_bc(int boxid, int bc) {
  set_bc( boxid, bc, bc, bc, bc, bc, bc);
}

//******************************************

void  FieldObj::set_bc(int boxid, const char *bcid) {
  set_bc( boxid, BCarray->bcidtoint(bcid) ); 
}

//************************************************************************************
//                              S E T   S O U R C E
//************************************************************************************

void  FieldObj::set_source(int id, double x, double y, double z, double sigmax, double sigmay, double sigmaz)
{

  if (VERBOSE) {
     fprintf(stdlog, "    Setting source #%d at (%g",id+1, x);
     if ((GEOMETRY & DIMENSION_MASK) > 1) fprintf(stdlog, ",%g",y);
     if ((GEOMETRY & DIMENSION_MASK) > 2) fprintf(stdlog, ",%g",z);
     fprintf(stdlog, "), with half-width of (%g", sigmax);
     if ((GEOMETRY & DIMENSION_MASK) > 1) fprintf(stdlog, ",%g",sigmay);
     if ((GEOMETRY & DIMENSION_MASK) > 2) fprintf(stdlog, ",%g",sigmaz);
     fprintf(stdlog, ")");
     if (weights[id] != 1.0) fprintf(stdlog, ", weight factor = %g", weights[id]);
     fprintf(stdlog, "\n");
  }

  if (id >= source_num)
    throw makeMessage("Bad source definition: source id=%d >= source_num=%d",id,source_num);

  if ( !Grid->in_bounds(x, y, z) ) {
    if ((GEOMETRY & DIMENSION_MASK) > 2) 
      throw makeMessage("Cannot initialize Ca source: location (%g,%g,%g) out of bounds",x,y,z);
    else if ((GEOMETRY & DIMENSION_MASK) > 1) 
      throw makeMessage("Cannot initialize Ca source: location (%g,%g) out of bounds",x,y);
    else  
      throw makeMessage("Cannot initialize Ca source: location (%g) out of bounds",x);
  }

  long ind = Grid->location_to_index(x, y, z);
  
  int ix, iy, iz;
  Grid->split(ind, ix, iy, iz);
  if (VERBOSE) {
     fprintf(stdlog, "    closest node: (%d", ix);
     if ((GEOMETRY & DIMENSION_MASK) > 1) fprintf(stdlog, ",%d",iy);
     if ((GEOMETRY & DIMENSION_MASK) > 2) fprintf(stdlog, ",%d",iz);
     fprintf(stdlog, ") <--> (%.4g", xcoord[ix]);
     if ((GEOMETRY & DIMENSION_MASK) > 1) fprintf(stdlog, ",%.4g",ycoord[iy]);
     if ((GEOMETRY & DIMENSION_MASK) > 2) fprintf(stdlog, ",%.4g",zcoord[iz]);
     fprintf(stdlog, ")\n");
  }

  //*************************

  if (VERBOSE > 2) { fprintf(stdlog,"      non-zero current on x-nodes "); fflush(stdlog); }
  set_source_xyz(id, x, ind, ix, sigmax, xcoord, xgrid, SURF_XMIN, SURF_XMAX, 1,
                 xsize - 1, source_x0, source_nx, source_xsum);

  if ( (GEOMETRY & DIMENSION_MASK) > 1) {
    if (VERBOSE > 2) { fprintf(stdlog,"      non-zero current on y-nodes "); fflush(stdlog); }
    set_source_xyz(id, y, ind, iy, sigmay, ycoord, ygrid, SURF_YMIN, SURF_YMAX, xsize,
                 ysize - 1, source_y0, source_ny, source_ysum);
  } else { 
    source_y0[id] = 0; source_ny[id] = 1;
    source_ysum[id] = new double[1]; source_ysum[id][0] = 1.0;
    }

  if ( (GEOMETRY & DIMENSION_MASK) > 2) {
    if (VERBOSE > 2)  { fprintf(stdlog,"      non-zero current on z-nodes "); fflush(stdlog); }
    set_source_xyz(id, z, ind, iz, sigmaz, zcoord, zgrid, SURF_ZMIN, SURF_ZMAX, xysize,
                 zsize - 1, source_z0, source_nz, source_zsum);
    }
  else { 
    source_z0[id] = 0; source_nz[id] = 1;
    source_zsum[id] = new double[1]; source_zsum[id][0] = 1.0;
    }
}

//************************************************************************************
//                            S O U R C E   S P R E A D
//************************************************************************************

void  FieldObj::set_source_xyz(int id, double c0, long i0, int ix, double sigma, 
               double *coord, double *grid, int minMask, int maxMask, int incr, int N,
               int *source_coord0, int *interval_num, double **source_weights )
{
if (sigma < 0.1 * grid[ix])  sigma = 0.1 * grid[ix];  // if source is point-like, set half-width at 
                                                      // 1/10th of grid spacing
  double *temp_weights = new double[N+1];
  double check_sum = 0.0;

  bool flag = false;    // to flag the entry into the region of non-zero current

  long i3d = i0;
  int  ileft = ix, iright = ix;
  while ( !( Region->point_type(i3d) & minMask) ) { i3d -= incr; ileft --; } 
  i3d = i0;
  while ( !( Region->point_type(i3d) & maxMask) ) { i3d += incr; iright ++; }

  for (int ii = ileft; ii <= iright; ii++) 
    {
    double dx0 = 0.5 * ( (ii > ileft)  ? grid[ii-1] : 0.0 );
    double dx1 = 0.5 * ( (ii < iright) ? grid[ii] : 0.0 );

    double x0 = coord[ii] - c0 - dx0;
    double x1 = coord[ii] - c0 + dx1;
    
    double left0 = 2 * (c0 - coord[ileft]) + x0;
    double left1 = 2 * (c0 - coord[ileft]) + x1;

    double right0 = 2 * (coord[iright] - c0) - x1;
    double right1 = 2 * (coord[iright] - c0) - x0;

    temp_weights[ii] = errf( x0, x1, sigma, 1.0 ) + 
                       errf( left0, left1, sigma, 1.0 ) + 
                       errf( right0, right1, sigma, 1.0 ); 
    double dv = (dx0 + dx1);  // default volume element

    if (GEOMETRY == SPHERICAL) {   // Spherical coordinates
      if (c0 == 0.0) temp_weights[ii] = 4 * pi * errf( x0, x1, sigma, 3.0 );
      dv = 4.0 / 3.0 * pi * ( pow( coord[ii] + dx1, 3.0 ) - pow( coord[ii] - dx0, 3.0 ) );
    }
    else if (GEOMETRY == CONICAL)  // Conical coordinates
      if (coord == xcoord) {
        if (c0 == 0.0) temp_weights[ii] = 4 * pi * errf( x0, x1, sigma, 3.0 );
        dv = 2 * pi / 3 * ( pow( coord[ii] + dx1, 3.0 ) - pow( coord[ii] - dx0, 3.0 ) );
      } else {
        if (c0 == 0.0) temp_weights[ii] = 2 * pi * errf( x0, x1, sigma, 2.0 );
        dv = cos( coord[ii] - dx0 ) - cos( coord[ii] + dx1 );
      }
    else if (GEOMETRY == CYLINDRICAL || GEOMETRY == DISC)  // Cylindrical coordinates 
      if (coord == xcoord) {
        if (c0 == 0.0) temp_weights[ii] = 2 * pi * errf( x0, x1, sigma, 2.0 );
        dv = pi * ( pow( coord[ii] + dx1, 2 ) - pow( coord[ii] - dx0, 2 ) );
      }
      // else (z direction) it's the default value of (dx0 + dx1)
    
    check_sum += temp_weights[ii];
    temp_weights[ii] /= dv;
   
    if ( temp_weights[ii] > SOURCE_EPS )   // integral between ii-1/2 and ii+1/2 > eps
      {
       if ( !flag )  
	 {
         flag = true;
         source_coord0[id] = ii;
         interval_num[id] = 1;
	 }
       else interval_num[id]++;
       if (VERBOSE > 2) fprintf(stdlog, "%d(%.5g):%.4g ", ii, coord[ii], temp_weights[ii]);
      }
    } // end for (ii = ileft; ii <= iright; ii++) 

  //*************************

  source_weights[id] = new double[interval_num[id]];

  for (int i = 0; i < interval_num[id]; i++) 
    source_weights[id][i] = temp_weights[i + source_coord0[id]] / check_sum;   // renormalize 

  if (fabs(check_sum - 1) > 1.0e-4) fprintf(stdlog,"*** WARNING: current total norm error=%.2e\n",check_sum - 1.0);

  if (VERBOSE > 2) fprintf(stdlog, "\n");

  delete [] temp_weights;
}

//************************************************************************************
//                      O L D    S O U R C E   S P R E A D
//************************************************************************************

double errf(double x0, double x1, double sigma, double dim)
{

  double eps = 1.0e-10;

  double a = 1.0 / ( pow(pi * sigma * sigma, 0.5 * dim) );
  double b = -1.0 / (sigma * sigma);

  long   N = 2;
  double resN = 0, res2N = 0, res4N = 0;
  double x, dx, sum;

  do
    {
    resN = res2N;
    N *= 2;
    dx = (x1 - x0) / double(N);
    sum = 0.0;
    for (int n = 1; n < N; n++)
      {
      x = x0 + dx * double(n);
      sum += ((n & 1) + 1) * exp( b * x * x ) * pow(x, dim-1);
      }
    res2N = dx / 3.0 * a * ( 2 * sum + exp( b * x0 * x0 ) * pow(x0, dim-1) + exp( b * x1 * x1 ) * pow(x1, dim-1) );
    res4N = (16.0 * res2N - resN) / 15.0;
    }
  while ( fabs(res4N - res2N) > eps * fabs(res4N + res2N) && fabs(res4N + res2N) > 1e-12);

  return res4N;
}


//***************************************************************

void  FieldObj::add_sources(VectorObj &f, double dt)
{
int    i, ix, iy, iz;
long   xind, yind, zind;
double addx, addy, charge;

 double ica = current->Evaluate();
 ICa = 0.0; 
 if (ica < MIN_CURRENT) return;
 double Q = ica * dt;

  for (i=0; i < source_num; i++) {

    charge = Q * weights[i];
    ICa += ica * weights[i];
    int nx = source_nx[i];
    int ny = source_ny[i];
    int nz = source_nz[i];
    xind = source_x0[i] + xsize * source_y0[i] + xysize * source_z0[i];
    //    fprintf(stderr,"%d %d %d  %d %d %d\n\n",nx,ny,nz, source_x0[i],source_y0[i],source_z0[i]);

    for (ix = 0; ix < nx; ix++, xind ++) {
      addx = charge * source_xsum[i][ix];
      for (iy = 0, yind = xind; iy < ny; iy++, yind += xsize) {
	addy = addx * source_ysum[i][iy];
        for (iz = 0, zind = yind; iz < nz; iz++, zind += xysize) 
           f.elem[zind] += addy * source_zsum[i][iz];
	   // fprintf(stderr,"%ld %g %g\n",zind, f.elem[zind], addy * source_zsum[i][iz]); }
      }
    }
  }

}

//************************************************************************
//            B O U N D A R Y   I N I T I A L I Z A T I O N
//************************************************************************
/*
   RegionObj::point_type(int,int,int) bit fields are:
   bit # is:      ...8       |   7    |   6  |  5   |  4   |  3   |  2   |  1  
   meaning:    box id number | inside | zmax | zmin | ymax | ymin | xmax | xmin

  ***************************************************************

   ptype array bit fields are:

   bit # is:  ...10    |    9    |    8      |   7    |   6  |  5   |  4   |  3   |  2   |  1  
   meaning:  bc_type # | outside | dirichlet | inside | zmax | zmin | ymax | ymin | xmax | xmin
  
   bit # is: |   30...24    |   23...17    |   16...10    |
   meaning:  | z bc_type id | y bc_type id | x bc_type id |

  ***************************************************************/

void FieldObj::init_boundary()
{
long      i;
int       p, bc = 0, boxid;
int       ix, iy, iz;

 for (i = 0; i < size; i++) ptype[i] =  _INSIDE_;   // default: point inside

 for (i = 0; i < size; i++)
   {
   Grid->split(i, ix, iy, iz);
   p = (*Region).point_type(ix, iy, iz); 
   boxid = p >> 7;
   p = p & INSIDE_MASK;              // remove box id (keep bits 1 through 7)
   if (p == 0)                       // point is outside
     {
     ptype[i] = _OUTSIDE_; 
     elem[i] = 0.0;
     }
   else if (p & SURF_MASK)           // point on the surface
     {
     if (p & SURF_XMIN)
       if (bc_deriv[bc = get_bc(boxid,0)] != 0) ptype[i] |= (bc << XSHIFT) + SURF_XMIN;
       else  { ptype[i] = (bc << XSHIFT) + DIRICHLET; continue; }
     
     if (p & SURF_XMAX) 
       if (bc_deriv[bc = get_bc(boxid,1)] != 0) ptype[i] |= (bc << XSHIFT) + SURF_XMAX;
          else { ptype[i] = (bc << XSHIFT) + DIRICHLET; continue; } 

     if (p & SURF_YMIN) 
       if (bc_deriv[bc = get_bc(boxid,2)] != 0) ptype[i] |= (bc << YSHIFT) + SURF_YMIN;
          else { ptype[i] = (bc << XSHIFT) + DIRICHLET; continue; }
     
     if (p & SURF_YMAX) 
       if (bc_deriv[bc = get_bc(boxid,3)] != 0) ptype[i] |= (bc << YSHIFT) + SURF_YMAX;
          else { ptype[i] = (bc << XSHIFT) + DIRICHLET; continue; }

     if (p & SURF_ZMIN)
       if (bc_deriv[bc = get_bc(boxid,4)] != 0) ptype[i] |= (bc << ZSHIFT) + SURF_ZMIN;
       else { ptype[i] = (bc << XSHIFT) + DIRICHLET; continue; }
     
     if (p & SURF_ZMAX)
       if (bc_deriv[bc = get_bc(boxid,5)] != 0) ptype[i] |= (bc << ZSHIFT) + SURF_ZMAX;
       else { ptype[i] = (bc << XSHIFT) + DIRICHLET; continue; }

     } // end if (p & SURF_MASK)
  
  } // end for loop


 for (i = 0; i < size; i++)   // SECOND PASS: take care of Dirichlet bc
   if ( ptype[i] & DIRICHLET ) 
     {
     elem[i] = bc_coef[ bc = ptype[i] >> XSHIFT ];
     Grid->split(i, ix, iy, iz);
     p = Region->point_type(ix, iy, iz); 
 
     if ( ix + 1 < xsize ) if ((Region->point_type(ix+1,iy,iz) & _INSIDE_) && (ptype[i+1] & VARY_MASK))
       ptype[i + 1] |= (bc << XSHIFT) + SURF_XMIN; 

     if ( ix - 1 > 0 )     if ((Region->point_type(ix-1,iy,iz) &  _INSIDE_) && (ptype[i-1] & VARY_MASK))
       ptype[i - 1] |= (bc << XSHIFT) + SURF_XMAX; 

     if ( iy + 1 < ysize ) if ((Region->point_type(ix,iy+1,iz) & _INSIDE_) && (ptype[i+xsize] & VARY_MASK))
       ptype[i + xsize] |= (bc << YSHIFT) + SURF_YMIN; 
     
     if ( iy - 1 > 0 )   if ((Region->point_type(ix,iy-1,iz) & _INSIDE_) && (ptype[i-xsize] & VARY_MASK))
       ptype[i - xsize] |= (bc << YSHIFT) + SURF_YMAX; 

     if ( iz + 1 < zsize ) if ((Region->point_type(ix,iy,iz+1) & _INSIDE_) && (ptype[i+xysize] & VARY_MASK))
        ptype[i + xysize] |= (bc << ZSHIFT) + SURF_ZMIN; 
     
     if ( iz - 1 > 0 ) if ((Region->point_type(ix,iy,iz-1) & _INSIDE_) && (ptype[i-xysize] & VARY_MASK))
        ptype[i - xysize] |= (bc << ZSHIFT) + SURF_ZMAX; 

  } // end for loop

}


//*****************************************************************************
/*
double FieldObj::interpolate(double x, double y, double z)
{
int ix = 0, iy = 0, iz = 0;

while (x > xcoord[ix] && ix < (xsize - 1) ) ix++;
while (y > ycoord[iy] && iy < (ysize - 1) ) iy++;
while (z > zcoord[iz] && iz < (zsize - 1) ) iz++;

long i = ix + xsize * iy + xysize * iz;

double p1 = (x - xcoord[ix-1]) / (xcoord[ix] - xcoord[ix-1]);
double p2 = (y - ycoord[iy-1]) / (ycoord[iy] - ycoord[iy-1]);
double p3 = (z - zcoord[iz-1]) / (zcoord[iz] - zcoord[iz-1]);

double res =      p1  *      p2  *      p3  * elem[i                     ] 
           + (1 - p1) * (1 - p2) * (1 - p3) * elem[i - 1 - xsize - xysize]
           +      p1  * (1 - p2) * (1 - p3) * elem[i     - xsize - xysize]
           + (1 - p1) *      p2  * (1 - p3) * elem[i - 1         - xysize]
           + (1 - p1) * (1 - p2) *      p3  * elem[i - 1 - xsize         ]
           +      p1  *      p2  * (1 - p3) * elem[i             - xysize]
           +      p1  * (1 - p2) *      p3  * elem[i     - xsize         ]
           + (1 - p1) *      p2  *      p3  * elem[i - 1                 ];

return res;
}
*/
//***********************************************************************************
/*
double FieldObj::get_value(double x, double y, double z)
{
return elem[location_to_index(x, y, z)];
}
*/
//***********************************************************************************

double FieldObj::sum()
{
double s = 0.0;
for(long i=0; i<size; i++) if (ptype[i] & REGION_MASK) s += elem[i];
return s;
}

//***********************************************************************************

double FieldObj::averageSpherical(double *V)
{
 double s = 0.0, v = 0.0, dv;

 for (int i = 1; i < xsize; i++ )

   if ( ! ((ptype[i] | ptype[i-1]) & _OUTSIDE_) )

     if ( xcoord[i-1] == 0.0 ) {                            // int( f(r) r^2, r=0..dr ) = 
       v += ( dv = 2 / 3.0 * pow( xcoord[i], 3 ) );         // int( ( f0 + r/dr (f1-f0) ) r^2, r=0..dr) =
       s += 3.0 / 4.0 * dv * (elem[i] + elem[i-1] / 3.0 );  // dr^3 ( f0/3 + (f1-f0)/4 ) = dr^3 / 4 ( f0/3 + f1 )
     }        
     else  {
       v += ( dv = xgrid[i-1] * (xcoord[i-1] * xcoord[i-1] + xcoord[i] * xcoord[i] ) );
       s += xgrid[i-1] * (elem[i-1] * xcoord[i-1] * xcoord[i-1] + elem[i] * xcoord[i] * xcoord[i]);
     }

if (V) *V = 2 * pi * v;
return s / v ;
}

//***********************************************************************************
/*
double FieldObj::averageConical(double *V)  // calculate integral( field[r,theta] sin(theta) r^2 dr dtheta )
{
 double s = 0.0, v = 0.0, Ith0, Ith1, Ir0, Ir1, rp, rm, thp, thm, dyInv, dvInv;

 for (int iy = 1; iy < ysize; iy++ )  {

   Ith0 = cos( ycoord[iy-1] ) - cos( ycoord[iy] );             // integral of sin(theta) dtheta
   Ith1 = sin( ycoord[iy] ) - sin( ycoord[iy-1] ) +            // integral of theta sin(theta) dtheta
          ycoord[iy-1] * Ith0 - ygrid[iy-1] * cos( ycoord[iy] );
   thp = Ith1 - ycoord[iy-1] * Ith0;
   thm = ycoord[iy] * Ith0 - Ith1;
   dyInv = 1 / ygrid[iy-1];

   for (int ix = 1; ix < xsize; ix++ ) {
     long i = ix + xsize * iy;
     if ( ! ((ptype[i] | ptype[i-1] | ptype[i-xsize] | ptype[i-1-xsize] ) & _OUTSIDE_) ) {

       Ir0 = ( pow( xcoord[ix], 3) - pow( xcoord[ix-1], 3) ) / 3.0;  // integral of r^2 dr
       Ir1 = ( pow( xcoord[ix], 4) - pow( xcoord[ix-1], 4) ) / 4.0;  // integral of r r^2 dr
       rp = Ir1 - xcoord[ix-1] * Ir0;
       rm = xcoord[ix] * Ir0 - Ir1;
       dvInv = dyInv / xgrid[ix-1];

       s += dvInv * ( elem[i-1-xsize] * rm * thm + elem[i-1] * rm * thp + 
                                          elem[i-xsize] * rp * thm + elem[i] * rp * thp );
       v += dvInv * ( rm + rp ) * ( thm + thp );
     }

   }
 }

if (V) *V = v;
return s / v ;
}
*/
//***********************************************************************************
 
double FieldObj::averageConical(double *V)
{
 double s = 0.0, v = 0.0, dv, dvy;

 for (int iy = 1; iy < ysize; iy++ )  {

   dvy = cos(ycoord[iy-1]) - cos(ycoord[iy]);

   for (int ix = 1; ix < xsize; ix++ ) {
     long i = ix + xsize * iy;
     if ( ! ((ptype[i] | ptype[i-1] | ptype[i-xsize] | ptype[i-1-xsize] ) & _OUTSIDE_) ) {

       v += ( dv = dvy * ( pow(xcoord[ix],3) - pow(xcoord[ix-1],3) ) );
       s += dv * (elem[i-1-xsize]  + elem[i-1] + elem[i-xsize] + elem[i]);
     }
   }
 }

if (V) *V = 2 * pi / 3.0 * v;
return 0.25 * s / v;
}

//***********************************************************************************
 
double FieldObj::averageCylindrical(double *V)
{
 double s = 0.0, v = 0.0, dv, dvy;

 for (int iy = 1; iy < ysize; iy++ )  {

   dvy = ygrid[iy-1];

   for (int ix = 1; ix < xsize; ix++ ) {
     long i = ix + xsize * iy;
     if ( ! ((ptype[i] | ptype[i-1] | ptype[i-xsize] | ptype[i-1-xsize] ) & _OUTSIDE_) ) {

       v += ( dv = dvy * ( xcoord[ix] * xcoord[ix] - xcoord[ix-1] * xcoord[ix-1] ) );
       s += dv * (elem[i-1-xsize]  + elem[i-1] + elem[i-xsize] + elem[i]);
     }
   }
 }

if (V) *V = pi * v;
return 0.25 * s / v;
}
//***********************************************************************************
 
double FieldObj::averageDisc(double *V)
{
 double s = 0.0, v = 0.0, dv;

   for (int i = 1; i < xsize; i++ ) 
     if ( ! ((ptype[i] | ptype[i-1]) & _OUTSIDE_) ) {
       v += ( dv = xcoord[i] * xcoord[i] - xcoord[i-1] * xcoord[i-1] );
       s += dv * (elem[i-1] + elem[i]);
     }

if (V) *V = pi * v;
return 0.5 * s / v;
}
//***********************************************************************************

double FieldObj::average3D(double *V)
{
double s = 0.0, v = 0.0, dv, dvx, dvxy;
int    ix, iy, iz;
long   i = 0;
double *e;

for(ix = 1; ix < xsize; ix++)
  {
  dvx = xgrid[ix-1];
  for(iy = 1; iy < ysize; iy++) 
    {
    dvxy = dvx * ygrid[iy-1];
    for(iz = 1; iz < zsize; iz++) 
      {
      e = elem + (i = ix + xsize * iy + xysize * iz);
  
      if ( ! ((ptype[i] | ptype[i-1] | ptype[i-xsize] | ptype[i-xysize] | ptype[i-1-xsize] |
	       ptype[i-xsize-xysize] | ptype[i-1-xysize] | ptype[i-1-xsize-xysize]) & _OUTSIDE_) )
	{
	  v += ( dv = dvxy * zgrid[iz-1] );

          s += dv * (*e +
           *(e-1) + *(e-xsize) + *(e-xysize) +
           *(e-1-xsize) + *(e-xsize-xysize) + *(e-1-xysize) +
           *(e-1-xsize-xysize) );  
	}
      }
    }
  }
 if (V) *V = v;
 return s / v * 0.125;
}

//***********************************************************************************

double FieldObj::average2D(double *V)
{
double s = 0.0, v = 0.0, dv, dvx;
int    ix, iy;
long   i = 0;
double *e;

  for(ix = 1; ix < xsize; ix++) 
    {
    dvx = xgrid[ix-1];
    for(iy = 1; iy < ysize; iy++) 
      {
      e = elem + (i = ix + xsize * iy);
  
      if ( ! ((ptype[i] | ptype[i-1] | ptype[i-xsize] | ptype[i-1-xsize]) & _OUTSIDE_) )
	{
	  v += ( dv = dvx * ygrid[iy-1] );
          s += dv * (*e + *(e-1) + *(e-xsize) + *(e-1-xsize) );  
	}
      }
    }
  
 if (V) *V = v;
 return s / v * 0.25;
}

//***********************************************************************************

double FieldObj::average1D(double *V)
{
double s = 0.0, v = 0.0, dv;
long   i = 0;

for(i = 1; i < xsize; i++)
  if ( ! ((ptype[i] | ptype[i-1]) & _OUTSIDE_) ) {
    v += (dv = xgrid[i-1]);
    s += dv * (elem[i] + elem[i-1]);
  }

 if (V) *V = v;
 return s / v * 0.5;
}

//***********************************************************************************
/*
double FieldObj::FluxBox(double *V)
{
double s = 0.0, v = 0.0, dv, dvx, dvxy;
int    ix, iy, iz;
long   i = 0;
double *e;

for(ix = 1; ix < xsize; ix++)
  {
  dvx = xgrid[ix-1];
  for(iy = 1; iy < ysize; iy++) 
    {
    dvxy = dvx * ygrid[iy-1];
    for(iz = 1; iz < zsize; iz++) 
      {
      e = elem + (i = ix + xsize * iy + xysize * iz);
      int p = ptype[i];
      if ( p & SURF_MASK ) {

        if ( ( p & SURF_XMIN ) 



      if ( ! ((ptype[i] | ptype[i-1] | ptype[i-xsize] | ptype[i-xysize] | ptype[i-1-xsize] |
	       ptype[i-xsize-xysize] | ptype[i-1-xysize] | ptype[i-1-xsize-xysize]) & _OUTSIDE_) )
	{
	  v += ( dv = dvxy * zgrid[iz-1] );

          s += dv * (*e +
           *(e-1) + *(e-xsize) + *(e-xysize) +
           *(e-1-xsize) + *(e-xsize-xysize) + *(e-1-xysize) +
           *(e-1-xsize-xysize) );  
	}
      }
    }
  }
 if (V) *V = v;
 return s / v * 0.125;
}
*/
//***********************************************************************************

double FieldObj::L2norm()
{
double s = 0.0;
for(long i=0; i<size; i++) if (ptype[i] & REGION_MASK) s += elem[i] * elem[i];
return s / double(size);
}


//***********************************************************************************
/*
void print3d(long *var)
{
for (int iz=0; iz < zsize; iz++, fprintf(stdlog,("\n"))
 for (int iy=0; iy < ysize; iy++, fprintf(stdlog,("\n"))
  for (int ix=0; ix < xsize; ix++) 
      fprintf(stdlog,("%ld ",var[ix + iy * xsize + iz * xysize]);
}
*/  
//***********************************************************************************
/*
void FieldObj::print3d(int n)
{
        char s1[20], s2[20];
        strcpy(s1,"%.");
        sprintf(s2,"%d",n);
        strcat(s1,s2);
        strcat(s1,"f ");

for (int iz=0; iz<zsize; iz++, fprintf(stdlog,"\n"))
 for (int iy=0; iy<ysize; iy++, fprintf(stdlog,"\n"))
  for (int ix=0; ix<xsize; ix++) fprintf(stdlog,s1,elem[ix+iy*xsize+iz*xysize]);

}
*/
//***********************************************************************************

void FieldObj::print_ptype()
{
for (int iz=0; iz < zsize; iz++, fprintf(stdlog,"\n"))
 for (int iy=0; iy < ysize; iy++, fprintf(stdlog,"\n"))
  for (int ix=0; ix < xsize; ix++) 
      print_ptype(ix + iy * xsize + iz * xysize);
}

//***************************************************************************************
//                              X   D I R E C T I O N
//***************************************************************************************
/*
 Find vector Result_new (to be stored in array Result) satisfying

 (1 + c * Dx^2 - Lin) Result_new = ( cx * Dx^2 + cy * Dy^2 + cz * Dz^2 ) This + Result 

 | diag[1] sup[1]   0      0 :                    |                | right[1] |
 | sub[2]  diag[2] sup[2]  0 :            0       |                |          |
 |...........................:....................| x Result_new = |          |
 |                           : diag[n-1] sup[n-1] |                |          |
 |            0              : sub[n]    diag[n]  |                | right[n] |

 Operator Dx^2:   Dx^2 f[i] == xminus[i] f[ix-1] + xcenter[i] f[i] + xplus[i] f[ix+1]
                          i == (ix,iy,iz)

***************************************************************************************/

void FieldObj::Run3Dx(double c, double cx, double cy, double cz, double *Lin, double *Result)
{
int      ix = 0, iy = 0, iz = 0;
long     i = 0, istart = 0, bcvar;
int      n = 0;
double   xminus, xplus, xcenter, xcoef;
double   yminus, yplus, ycenter, ycoef;
double   zminus, zplus, zcenter, zcoef;
double   ri;


if ( D <= 0.0 )  {
  for (i = 0; i < size; i ++)  if (ptype[i] & VARY_MASK) 
      Result[i] /= (1.0 -  (Lin ? Lin[i] : 0.0) + 1.0e-16);
  return;
}

do  {
  while ( (bcvar = ptype[i]) & VARY_MASK )  {
    n++;
    if ( bcvar & SURF_XMIN ) istart = i; 
    nabla1Dx(i, ix, bcvar, xminus, xcenter, xplus, xcoef);
    nabla1Dy(i, iy, bcvar, yminus, ycenter, yplus, ycoef);
    nabla1Dz(i, iz, bcvar, zminus, zcenter, zplus, zcoef);
    
    ri = Result[i] + (cx - c) * xcoef + cy * ycoef + cz * zcoef + (cx * xcenter + cy * ycenter + cz * zcenter) * elem[i];
    if ( ! (bcvar & SURF_XMIN) )  ri += cx * xminus * elem[i-1]; 
    if ( ! (bcvar & SURF_XMAX) )  ri += cx * xplus  * elem[i+1];     
    if ( ! (bcvar & SURF_YMIN) )  ri += cy * yminus * elem[i-xsize]; 
    if ( ! (bcvar & SURF_YMAX) )  ri += cy * yplus  * elem[i+xsize];    
    if ( ! (bcvar & SURF_ZMIN) )  ri += cz * zminus * elem[i-xysize]; 
    if ( ! (bcvar & SURF_ZMAX) )  ri += cz * zplus  * elem[i+xysize];    
 
    diag[n] = 1 - (Lin ? Lin[i] : 0.0) + c * xcenter;
    sup[n] = c * xplus;  sub[n] = c * xminus;
    right[n] = ri; 

    i ++;
    if (++ix == xsize) break;
  } // while bcvar > 0 and ix < xsize

  if (n)  { 
    tridiag(n);
    for (int ii = 1, ind = istart; ii<=n; ii++, ind ++) Result[ind] = right[ii];
    n = 0;
  }

  if (ix < xsize) 
    while ( !(ptype[i] & VARY_MASK) ) {
      if ( ptype[i] & DIRICHLET ) Result[i] = bc_coef[ptype[i] >> XSHIFT];
      i ++; 
      if ( (++ix) == xsize ) break; 
    }

  if (ix == xsize) { 
     ix = 0;
     if (++iy == ysize) { iy = 0; iz ++; } 
  }
  
  } while (iz < zsize);   
}

//***************************************************************************************
//                               Y   D I R E C T I O N
//
//  (1 + c * Dy^2 - Lin) Result_new = ( cx * Dx^2 + cy * Dy^2 + cz * Dz^2 ) This + Result
//
//***************************************************************************************

void FieldObj::Run3Dy(double c, double cx, double cy, double cz, double *Lin, double *Result)
{
int      ix = 0, iz = 0, iy = 0;
long     i = 0, istart = 0, bcvar;
int      n = 0;
double   xminus, xplus, xcenter, xcoef;
double   yminus, yplus, ycenter, ycoef;
double   zminus, zplus, zcenter, zcoef;
double   ri;

if ( D <= 0.0 )  {
  for (i = 0; i < size; i ++) if (ptype[i] & VARY_MASK) 
     Result[i] /= (1.0 -  (Lin ? Lin[i] : 0.0) + 1.0e-12);
  return;
}

do  {
  while ( (bcvar = ptype[i]) & VARY_MASK ) {
    n++;
    if ( bcvar & SURF_YMIN ) istart = i; 
    nabla1Dx(i, ix, bcvar, xminus, xcenter, xplus, xcoef);
    nabla1Dy(i, iy, bcvar, yminus, ycenter, yplus, ycoef);
    nabla1Dz(i, iz, bcvar, zminus, zcenter, zplus, zcoef);
    
    ri = Result[i] + (cy - c) * ycoef + cx * xcoef + cz * zcoef + (cx * xcenter + cy * ycenter + cz * zcenter) * elem[i];
    if ( ! (bcvar & SURF_XMIN) )  ri += cx * xminus * elem[i-1]; 
    if ( ! (bcvar & SURF_XMAX) )  ri += cx * xplus * elem[i+1];     
    if ( ! (bcvar & SURF_YMIN) )  ri += cy * yminus * elem[i-xsize]; 
    if ( ! (bcvar & SURF_YMAX) )  ri += cy * yplus * elem[i+xsize];    
    if ( ! (bcvar & SURF_ZMIN) )  ri += cz * zminus * elem[i-xysize]; 
    if ( ! (bcvar & SURF_ZMAX) )  ri += cz * zplus * elem[i+xysize];    

    diag[n] = 1 - (Lin ? Lin[i] : 0.0) + c * ycenter;
    sup[n] = c * yplus;  sub[n] = c * yminus;
    right[n] = ri; 

    i += xsize;  
    if (++iy == ysize) break;
  } // while bcvar > 0 and iy < ysize

  if (n > 0)  {
      tridiag(n);
      for (int ii = 1, ind = istart; ii<=n; ii++, ind += xsize) Result[ind] = right[ii]; 
      n = 0;
  }

  if (iy < ysize)
      while ( !(ptype[i] & VARY_MASK) ) {  
         if ( ptype[i] & DIRICHLET ) Result[i] = bc_coef[ptype[i] >> XSHIFT];
         i += xsize; if ( (++iy) == ysize ) break; 
  }

  if (iy == ysize)    {
      iy = 0 ;
      if (++iz == zsize) { i -= size - 1; iz = 0; ix++; }
  }

} while (ix < xsize); 

}


//***************************************************************************************
//                              Z   D I R E C T I O N
//***************************************************************************************
//
// (1 + c * Dz^2 - Lin) Result_new = ( cx * Dx^2 + cy * Dy^2 + cz * Dz^2 ) This + Result 
//
//***************************************************************************************

void FieldObj::Run3Dz(double c, double cx, double cy, double cz, double *Lin, double *Result)

{
int      iz = 0, ix = 0, iy = 0;
long     i = 0, istart = 0, bcvar;
int      n = 0;
double   xminus, xplus, xcenter, xcoef;
double   yminus, yplus, ycenter, ycoef;
double   zminus, zplus, zcenter, zcoef;
double   ri;

if ( D <= 0.0 )  {
  for (i = 0; i < size; i ++) if (ptype[i] & VARY_MASK) 
    Result[i] /= (1.0 - (Lin ? Lin[i] : 0.0) + 1.0e-12);
  return;
}

do  {
  while ( (bcvar = ptype[i]) & VARY_MASK ) {  
    n++;

    if ( bcvar & SURF_ZMIN ) istart = i; 
    nabla1Dx(i, ix, bcvar, xminus, xcenter, xplus, xcoef);
    nabla1Dy(i, iy, bcvar, yminus, ycenter, yplus, ycoef);
    nabla1Dz(i, iz, bcvar, zminus, zcenter, zplus, zcoef);
    
    ri = Result[i] + (cz - c) * zcoef + cx * xcoef + cz * zcoef + (cx * xcenter + cy * ycenter + cz * zcenter) * elem[i];
    if ( ! (bcvar & SURF_XMIN) )  ri += cx * xminus * elem[i-1]; 
    if ( ! (bcvar & SURF_XMAX) )  ri += cx * xplus * elem[i+1];     
    if ( ! (bcvar & SURF_YMIN) )  ri += cy * yminus * elem[i-xsize]; 
    if ( ! (bcvar & SURF_YMAX) )  ri += cy * yplus * elem[i+xsize];    
    if ( ! (bcvar & SURF_ZMIN) )  ri += cz * zminus * elem[i-xysize]; 
    if ( ! (bcvar & SURF_ZMAX) )  ri += cz * zplus * elem[i+xysize];    

    diag[n] = 1 - (Lin ? Lin[i] : 0.0) + c * zcenter;
    sup[n] = c * zplus;  sub[n] = c * zminus;
    right[n] = ri; 

    i += xysize; ++iz;
    if (iz == zsize) break;

  } // while ptype > 0 and iz < zsize

  if (n > 0)    {  
    tridiag(n);
    for (int ii = 1, ind = istart; ii<=n; ii++, ind += xysize) Result[ind] = right[ii]; 
    n = 0;
  } 
 
  if (iz < zsize)
    while ( !(ptype[i] & VARY_MASK) ) {
       if ( ptype[i] & DIRICHLET ) Result[i] = bc_coef[ptype[i] >> XSHIFT];
       i += xysize; if ( (++iz) == zsize) break; 
    }

  if (iz == zsize)  {
      iz = 0; i -= (size - 1);
      if (++ix == xsize) { ix = 0; iy ++; }
  }

} while (iy < ysize);

}

/***************************************************************************************
                                    1 D    R U N
                                    ^^^^^^^^^^^^

 Find vector Result_new (to be stored in array Result) satisfying

 (1 + c * Dx^2 - Lin) Result_new = cc * Dx^2 This + Result 

 | diag[1] sup[1]   0      0 :                    |                | right[1] |
 | sub[2]  diag[2] sup[2]  0 :            0       |                |          |
 |...........................:....................| x Result_new = |          |
 |                           : diag[n-1] sup[n-1] |                |          |
 |            0              : sub[n]    diag[n]  |                | right[n] |

 Operator Dx^2:   Dx^2 f[i] == minus f[i-1] + center f[i] + plus f[i+1] + coef
 off the boundary: minus = dxminus[i], center = dxcenter[i], plus = dxplus[i]

 sub[i]   = c minus
 diag[i]  = 1 + c center - Lin[i]
 sup[i]   = c plus
 right[i] = Result[i] + cc ( elem[i] center + elem[i-1] minus + elem[i+1] plus ) + (cc - c) coef
                                       
***************************************************************************************/

void FieldObj::Run1D( double c, double cc, double *Lin, double *Result)
{
long     i = 0, istart = 0, bcvar;
int      n = 0;
double   ri, plus, minus, center, coef;

if ( D <= 0.0 )  {
  for (i = 0; i < size; i ++)
    if (ptype[i] & VARY_MASK) Result[i] /= (1.0 -  (Lin ? Lin[i] : 0.0) + 1.0e-16);
  return;
  }

do  {
  while ( (bcvar = ptype[i]) & VARY_MASK ) {
    n++;
    if ( bcvar & SURF_XMIN )   istart = i; 

    (this->*nablaX)(i, i, bcvar, minus, center, plus, coef);

    ri = Result[i] + (cc - c) * coef + cc * center * elem[i];

    if ( ! (bcvar & SURF_XMIN) ) ri += cc * minus * elem[i-1];
    if ( ! (bcvar & SURF_XMAX) ) ri += cc * plus * elem[i+1]; 

    diag[n] = 1 - (Lin ? Lin[i] : 0.0) + c * center ;
    sup[n] = c * plus;  sub[n] = c * minus;
    right[n] = ri;
    if (++i == xsize) break;

    } // while bcvar > 0 and i < xsize
 
  if (n)  {
    tridiag(n);
    for (int ii = 1, ind = istart; ii<=n; ii++, ind ++) Result[ind] = right[ii];
    n = 0;
  }

  if (i < xsize)
    while ( !(ptype[i] & VARY_MASK) )   {
       if ( ptype[i] & DIRICHLET ) Result[i] = bc_coef[ptype[i] >> XSHIFT];
       if ( ++i == xsize) break; 
    }

  } while (i < xsize);

}

     
/***************************************************************************************
 
                     CYLINDRICAL GEOMETRY RUN, "r" DIRECTION
                                                            
 Solving  (1 + c Dr^2 - Lin) Result_new = ( cr Dr^2 + cz Dz^2 ) This + Result
                                                          
***************************************************************************************

                     CONICAL GEOMETRY RUN, "r" DIRECTION:
                                                           1 
 Solving  (1 + c Dr^2 - Lin) Result_new = ( cr Dr^2 + cth --- Dth^2 ) This + Result
                                                          r^2
***************************************************************************************/

void FieldObj::Run2Dx( double c, double cr, double cz, double *Lin, double *Result)
{
long     i = 0, istart = 0, bcvar;
int      ix = 0, iy = 0, n = 0;
double   a = 1.0, ri, rplus, rminus, rcenter, rcoef;
double   zplus, zminus, zcenter, zcoef;

if ( D <= 0.0 )  {
  for (i = 0; i < size; i ++)
    if (ptype[i] & VARY_MASK) Result[i] /= (1.0 - (Lin ? Lin[i] : 0.0) + 1.0e-16);
  return;
}

do  {
  while ( (bcvar = ptype[i]) & VARY_MASK )   {
    n++;
    if ( bcvar & SURF_XMIN )   istart = i; 
    (this->*nablaX)(i, ix, bcvar, rminus, rcenter, rplus, rcoef);     // calculating D_r^2
    (this->*nablaY)(i, iy, bcvar, zminus, zcenter, zplus, zcoef);

    if (GEOMETRY == CONICAL) {
      if (xcoord[ix] == 0) a = 0.0;   // no nabla theta at origin
      else a = 1 / ( xcoord[ix] * xcoord[ix] ); 
    }

    ri = Result[i] + (cr - c) * rcoef + a * cz * zcoef + (cr * rcenter + a * cz * zcenter) * elem[i];
    if ( ! (bcvar & SURF_XMIN) )  ri += cr * rminus * elem[i-1]; 
    if ( ! (bcvar & SURF_XMAX) )  ri += cr * rplus * elem[i+1];     
    if ( ! (bcvar & SURF_YMIN) )  ri += a * cz * zminus * elem[i-xsize]; 
    if ( ! (bcvar & SURF_YMAX) )  ri += a * cz * zplus * elem[i+xsize];    

    diag[n] = 1 - (Lin ? Lin[i] : 0.0) + c * rcenter;
    sup[n] = c * rplus;  sub[n] = c * rminus;
    right[n] = ri; 
    ++i;
    if (++ix == xsize) break;

  } // while bcvar > 0 and i < xsize
 
  if (n)  {
    tridiag(n);
    for (int ii = 1, ind = istart; ii<=n; ii++, ind ++) Result[ind] = right[ii];
    n = 0;
  }

  if (ix < xsize)
    while ( !(ptype[i] & VARY_MASK) )   {
       if ( ptype[i] & DIRICHLET ) Result[i] = bc_coef[ptype[i] >> XSHIFT];
       ++i;
       if ( ++ix == xsize) break; 
    }

  if (ix == xsize) {
    ix = 0;
    iy ++;
  }

} while (iy < ysize);

}

/***************************************************************************************
 
                     CYLINDRICAL GEOMETRY RUN, "z" DIRECTION
                                                            
 Solving  (1 + c Dz^2 - Lin) Result_new = ( cr Dr^2 + cz Dz^2 ) This + Result
                                                          
***************************************************************************************

                     CONICAL GEOMETRY RUN, "theta" DIRECTION:
                  1                                             1 
 Solving  (1 + c --- Dth^2 - Lin) Result_new = ( cr Dr^2 + cth --- Dth^2 ) This + Result
                 r^2                                           r^2

***************************************************************************************/

void FieldObj::Run2Dy( double c, double cr, double cz, double *Lin, double *Result)
{
long     i = 0, istart = 0, bcvar;
int      ix = 0, iy = 0, n = 0;
double   ri, a = 1.0;
double   rplus, rminus, rcenter, rcoef;
double   zplus, zminus, zcenter, zcoef;

if ( D <= 0.0 )  {
  for (i = 0; i < size; i ++)
    if (ptype[i] & VARY_MASK) Result[i] /= (1.0 - (Lin ? Lin[i] : 0.0) + 1.0e-16);
  return;
}

do  {
  while ( (bcvar = ptype[i]) & VARY_MASK ) {

    n++;
    if ( bcvar & SURF_YMIN )  istart = i; 

    (this->*nablaY)(i, iy, bcvar, zminus, zcenter, zplus, zcoef);
    (this->*nablaX)(i, ix, bcvar, rminus, rcenter, rplus, rcoef);     // calculating D_r^2  

    if (GEOMETRY == CONICAL) {
         if (xcoord[ix] == 0) a = 0.0;   // no nabla theta at origin
         else a = 1 / ( xcoord[ix] * xcoord[ix] ); 
    }
   
    ri = Result[i] + cr * rcoef + (cz - c) * a * zcoef + (cr * rcenter + a * cz * zcenter) * elem[i];
    if ( ! (bcvar & SURF_XMIN) ) ri += cr * rminus * elem[i-1];
    if ( ! (bcvar & SURF_XMAX) ) ri += cr * rplus * elem[i+1];   
    if ( ! (bcvar & SURF_YMIN) ) ri += a * cz * zminus * elem[i-xsize];
    if ( ! (bcvar & SURF_YMAX) ) ri += a * cz * zplus * elem[i+xsize];   

    diag[n] = 1 - (Lin ? Lin[i] : 0.0) + a * c * zcenter;
    sup[n] = a * c * zplus; sub[n] = a * c * zminus;
    right[n] = ri; 

    i += xsize;
    if (++iy == ysize) break;

    } // while bcvar > 0 and iy < ysize
 
  if (n)
    {
      //for  (int t = 1; t <= n; t++) fprintf(stderr,"%g %g %g %g\n",sub[t],diag[t],sup[t],right[t]);
    tridiag(n);
    for (int ii = 1, ind = istart; ii<=n; ii++, ind += xsize) Result[ind] = right[ii];
    n = 0;
    }

  if (iy < ysize)
    while ( !(ptype[i] & VARY_MASK) )  
       {
       if ( ptype[i] & DIRICHLET ) Result[i] = bc_coef[ptype[i] >> XSHIFT];
       i += xsize;
       if ( ++iy == ysize) break; 
       }

  if (iy == ysize) {
    iy = 0;
    i = ++ix;
  }

  } while (ix < xsize);

}

/***************************************************************************************

 Operator Dx^2:   Dx^2 f[i] == dxminus[i] f[ix-1] + dxcenter[i] f[i] + dxplus[i] f[ix+1]
                          i == (ix,iy,iz)

 Special case of Neumann boundary ( df/dx[i] + bc_lin f[i] = bc_coef ):
 Assume uniform for accuracy!!

  |       |      |    f[1] - f[-1] = 2 xgrid[0] ( bc_coef - f[0] bc_lin )
virtual f[0]   f[1]   Eliminating f[-1]: f[-1] = f[1] - 2 xgrid[0] ( bc_coef - f[0] bc_lin )
 node

  Dx^2 f[0] = ( -2 f[0] + f[1] + f[-1] ) / xgrid[0]^2
            = 2 ( f[1] - f[0] ) / xgrid[0]^2 - 2 ( bc_coef - f[0] bc_lin ) / xgrid[0]
            = 2 f[0] ( bc_lin - 1 / xgrid[0] ) / xgrid[0] +          | -> diag[0]
              2 f[1] / xgrid[0]^2 -                                  | -> sup[0]
            - 2 bc_coef / xgrid[0]                                   | -> right[0]

=======================================================================================

"Tortuosity": D = D(r)


             ( D[i+1/2] (f[i+1] - f[i]) / xgrid[i] - D[i-1/2] (f[i] - f[i-1]) / xgrid[i-1] ) 
 Dx^2 f[i] = ------------------------------------------------------------------------------- =
                                 (xgrid[i] + xgrid[i-1])  / 2

                    D[i] + D[i-1]                               D[i+1] + D[i]
 = (f[i-1] - f[i]) -------------- dxminus[i] + (f[i+1] - f[i]) -------------- dxplus[i]
                         2                                           2


Neumann b.c.:

  Dx^2 f[0] = ( D[1/2] ( f[1] - f[0] ) - D[-1/2] ( f[0] - f[-1] ) ) / xgrid[0]^2
            = ( 2 D[0] ( f[1] - f[0] ) / xgrid[0]^2 - 2 D[-1/2] ( bc_coef - f[0] bc_lin ) / xgrid[0]
            = 2 f[0] ( D[-1/2] bc_lin - D[0] / xgrid[0] ) / xgrid[0] +  | -> diag[0]
              2 D[0] f[1] / xgrid[0]^2 -                                | -> sup[0]
            - 2 D[-1/2] bc_coef / xgrid[0]                              | -> right[0]

  2 D[-1/2] = 3 D[0] - D[1]

***************************************************************************************/

void FieldObj::nabla1Dx(long i, int ix, long bcvar,
                      double &minus, double &center, double &plus, double &coef)
{
    int bc;
    double Di = Diff[i];

    if ( bcvar & SURF_XMIN ) {
      minus = 0;

      if ( bc_deriv[bc = (bcvar >> XSHIFT) & BC_ID_MASK ] == 0) 
	center = - (plus = 0.5 * (Di + Diff[i+1]) * dxplus[ix])
                 - (coef = 0.5 * (Di + Diff[i-1]) * dxminus[ix]);
      else 
        center = - (plus = 2 * Di * dx2inv[ix] )
                 - bc_lin[bc] * (coef = dxinv[ix] * (Di - 3 * Diff[i+1]));

      coef *= bc_coef[bc];
    }
    else if ( bcvar & SURF_XMAX ) {
      plus = 0;

      if ( bc_deriv[bc = (bcvar >> XSHIFT) & BC_ID_MASK ] == 0) 
	center = - (minus = 0.5 * (Di + Diff[i-1]) * dxminus[ix])
                 - (coef  = 0.5 * (Di + Diff[i+1]) * dxplus[ix]);
      else  
        center = - (minus = 2 * Di * dx2inv[ix-1] )
                 - bc_lin[bc] * (coef = dxinv[ix-1] * (Di - 3 * Diff[i-1]));

      coef *= bc_coef[bc];
    }
    else {
      center = - (plus  = 0.5 * (Di + Diff[i+1]) * dxplus[ix])
	       - (minus = 0.5 * (Di + Diff[i-1]) * dxminus[ix]);
      coef = 0.0;
    }
}
//******************* calculating D_y^2  *************************

void FieldObj::nabla1Dy(long i, int iy, long bcvar,
                      double &minus, double &center, double &plus, double &coef)
{
    int bc;
    double Di = Diff[i];

    if ( bcvar & SURF_YMIN ) {
      minus = 0;

      if ( bc_deriv[bc = (bcvar >> YSHIFT) & BC_ID_MASK ] == 0) 
	center = - (plus = 0.5 * (Di + Diff[i+xsize]) * dyplus[iy])
                 - (coef = 0.5 * (Di + Diff[i-xsize]) * dyminus[iy]);
      else 
        center = - (plus = 2 * Di * dy2inv[iy] )
                 - bc_lin[bc] * (coef = dyinv[iy] * (Di - 3 * Diff[i+xsize]));

      coef *= bc_coef[bc];
    }
    else if ( bcvar & SURF_YMAX ) {
      plus = 0;

      if ( bc_deriv[bc = (bcvar >> YSHIFT) & BC_ID_MASK ] == 0) 
	center = - (minus = 0.5 * (Di + Diff[i-xsize]) * dyminus[iy])
                 - (coef  = 0.5 * (Di + Diff[i+xsize]) * dyplus[iy]);
      else  
        center = - (minus = 2 * Di * dy2inv[iy-1] )
                 - bc_lin[bc] * (coef = dyinv[iy-1] * (Di - 3 * Diff[i-xsize]));

      coef *= bc_coef[bc];
      //fprintf(stderr,"YMAX: iy=%d bc=%d (%g,%g,%g) -> (%g,%g,%g,%g)\n",
      //        iy,bc,bc_deriv[bc],bc_lin[bc],bc_coef[bc],minus,center,plus,coef);
    }
    else {
      center = - (plus  = 0.5 * (Di + Diff[i+xsize]) * dyplus[iy])
	       - (minus = 0.5 * (Di + Diff[i-xsize]) * dyminus[iy]);
      coef = 0.0;
    }
}
//******************* calculating D_z^2  *************************

void FieldObj::nabla1Dz(long i, int iz, long bcvar,
                      double &minus, double &center, double &plus, double &coef)
{
    int bc;
    double Di = Diff[i];

    if ( bcvar & SURF_ZMIN ) {
      minus = 0;

      if ( bc_deriv[bc = (bcvar >> ZSHIFT) & BC_ID_MASK ] == 0) 
	center = - (plus = 0.5 * (Di + Diff[i+xysize]) * dzplus[iz])
                 - (coef = 0.5 * (Di + Diff[i-xysize]) * dzminus[iz]);
      else 
        center = - (plus = 2 * Di * dz2inv[iz] )
                 - bc_lin[bc] * (coef = dzinv[iz] * (Di - 3 * Diff[i+xysize]));

      coef *= bc_coef[bc];
    }
    else if ( bcvar & SURF_ZMAX ) {
      plus = 0;

      if ( bc_deriv[bc = (bcvar >> ZSHIFT) & BC_ID_MASK ] == 0) 
	center = - (minus = 0.5 * (Di + Diff[i-xysize]) * dzminus[iz])
                 - (coef  = 0.5 * (Di + Diff[i+xysize]) * dzplus[iz]);
      else  
        center = - (minus = 2 * Di * dz2inv[iz-1] )
                 - bc_lin[bc] * (coef = dzinv[iz-1] * (Di - 3 * Diff[i-xysize]));

      coef *= bc_coef[bc];
    }
    else {
      center = - (plus  = 0.5 * (Di + Diff[i+xysize]) * dzplus[iz])
	       - (minus = 0.5 * (Di + Diff[i-xysize]) * dzminus[iz]);
      coef = 0.0;
    }
}
/***************************************************************************************
 
                          POLAR COORDINATES DIFFERENTIAL 
                          ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^                                  
 with Tortuosity:

                1  d        df
 Dr^2 f[i,j] == - --- ( D r -- ) == rminus[i] f[i-1] + rcenter[i] f[i] + rplus[i] f[i+1]
                r  dr       dr 
 where r == r[i]

  r[i+1/2] D[i+1/2] (f[i+1]-f[i]) / dr[i] - D[i-1/2] r[i-1/2] (f[i] - f[i-1]) / dr[i-1] 
= ------------------------------------------------------------------------------------- =
                           r[i]  ( dr[i] + dr[i-1] ) / 2

         2                      D[i+1/2] r[i+1/2]          D[i-1/2] r[i-1/2]         
---------------------- { f[i+1] ----------------- + f[i-1] ----------------- - f[i] (..sum..) } =
r[i] (dr[i] + dr[i-1])                dr[i]                      dr[i-1]            


 rplus[i]  = (D[i] + D[i+1]) / 2 * ( 2 / dr[i]   + 1 / r[i] ) / (dr[i] + dr[i-1] ) 
 rminus[i] = (D[i] - D[i+1]) / 2 * ( 2 / dr[i-1] - 1 / r[i] ) / (dr[i] + dr[i-1] ) 

----------------------------------

 Special case of Neumann boundary ( df/dx[i] + bc_lin f[i] = bc_coef ):
 Assume uniform grid at boundary for accuracy!!

  |       |      |    f[1] - f[-1] = 2 xgrid[0] ( bc_coef - f[0] bc_lin )
virtual f[0]   f[1]   Eliminating f[-1]: f[-1] = f[1] - 2 dr ( bc_coef - f[0] bc_lin )
 node

               1                  r + dr/2                 r - dr/2         
 Dr^2 f[0] = ------ { D[1/2] f[1] -------- + D[-1/2] f[-1] -------- - f[0] ( ..sum.. ) }
             r dr                    dr                       dr            

 D[-1/2] = (3 D[0] - D[1]) / 2

   1    /               / 2 r        D[1]-D[0] \                    dr                             \
= ---- | (f[1] - f[0]) | --- D[0] + ---------   | - 2 D[-1/2] ( r - --- ) (bc_coef - bc_lin * f[0]) |
  r dr  \               \ dr             2     /                     2                             /

                   /  2          D[1]-D[0] \              2    1 
  = (f[1] - f[0]) | ----- D[0] + --------   | - D[-1/2] ( -- - - ) ( bc_coef - bc_lin * f[0] )
                   \ dr^2         2 r dr   /              dr   r

***************************************************************************************/

void FieldObj::nablaDiscR(long i, int ix, long bcvar, 
                      double &rminus, double &rcenter, double &rplus, double &rcoef)
{
  double a, b;
  int bc;

    rcenter = - ( (rminus = dxminus[ix]) + (rplus = dxplus[ix]) );
    rcoef = 0.0;

    if ( bcvar & SURF_XMIN )
      {
      rminus = 0.0;
      if ( bc_deriv[bc = (bcvar >> XSHIFT) & BC_ID_MASK ] == 0)  
        rcoef = dxminus[ix] * bc_coef[bc];
      else 
        if (xcoord[ix] > 0 ) { 
          a = 2 * dxinv[ix];                   // a = 2 / dr
          b = a - 1 / xcoord[ix];              // b = 2/dr - 1 / r
          rplus = a * dxinv[ix];               // rplus = 2 / dr^2
          rcoef = - b * bc_coef[bc];           // rcoef = - (2/dr - 1/r) Coef
          rcenter = - rplus + b * bc_lin[bc];  // rcenter = - 2 / dr^2 + Lin (2 / dr - 1 /r)
 	  }
        else rcenter = - ( rplus = 4 * dx2inv[ix] );
      }
    else if ( bcvar & SURF_XMAX ) 
      {
      rplus = 0.0;
      if ( bc_deriv[bc = (bcvar >> XSHIFT) & BC_ID_MASK ] == 0)  
         rcoef = dxplus[ix] * bc_coef[bc];
      else 
        {
        a = 2 * dxinv[ix-1];                 // a = 2 / dr
        b = a + 1 / xcoord[ix];              // b = 2/dr + 1 / r
        rminus = a * dxinv[ix-1];             // rplus = 2 / dr^2
        rcoef = - b * bc_coef[bc];           // rcoef = - (2/dr + 1/r) Coef
        rcenter = - rminus + b * bc_lin[bc];  // rcenter = - 2 / dr^2 + Lin (2 / dr - 1 /r)
        }
      }
}
/***************************************************************************************

                       SPHERICAL COORDINATES DIFFERENTIAL
                       ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

                 1   d        df
 Dr^2 f[i,j] == --- --- ( r^2 -- ) == rminus[i] f[i-1] + rcenter[i] f[i] + rplus[i] f[i+1]
                r^2  dr       dr 
 where r == r[i]

       1   r[i+1/2]^2 (f[i+1]-f[i]) / dr[i] - r[i-1/2]^2 (f[i] - f[i-1]) / dr[i-1] 
 =  ------ ----------------------------------------------------------------------- =
    r[i]^2                          ( dr[i] + dr[i-1] ) / 2

           2                       r[i+1/2]^2          r[i-1/2]^2          r[i+1/2]^2   r[i-1/2]^2
 ------------------------ { f[i+1] ---------- + f[i-1] ---------- - f[i] ( ---------- + ---------- ) }
 r[i]^2 (dr[i] + dr[i-1])             dr[i]              dr[i-1]              dr[i]       dr[i-1]

----------------------------------

 Special case of Neumann boundary ( df/dx[i] + bc_lin f[i] = bc_coef ):
 Assume uniform grid at boundary for accuracy!!

  |       |      |    f[1] - f[-1] = 2 dr ( bc_coef - f[0] bc_lin )
virtual f[0]   f[1]   Eliminating f[-1]: f[-1] = f[1] - 2 dr ( bc_coef - f[0] bc_lin )
 node

               1           (r + dr/2)^2         (r - dr/2)^2          (r + dr/2)^2   (r - dr/2)^2
 Dr^2 f[0] = ------ { f[1] ------------ + f[-1] ------------ - f[0] ( ------------ + ------------ ) }
             r^2 dr             dr                  dr                     dr             dr

      1                           4 r^2               2 r
  = ----- { ( f[1] - f[0] ) ( 1 + ------ ) - dr ( 1 - ---- )^2 ( bc_coef - bc_lin * f[0] ) }
    2 r^2                          dr^2                dr

-----------------------------------

  |       |      |      f[-1] - f[1] = 2 dr ( bc_coef - f[0] bc_lin )
 f[-1]  f[0]  virtual   Eliminating f[1]: f[1] = f[-1] - 2 dr ( bc_coef - f[0] bc_lin )
               node

               1           (r + dr/2)^2         (r - dr/2)^2          (r + dr/2)^2   (r - dr/2)^2
 Dr^2 f[0] = ------ { f[1] ------------ + f[-1] ------------ - f[0] ( ------------ + ------------ ) }
             r^2 dr             dr                  dr                     dr             dr

      1                            4 r^2               2 r
  = ----- { ( f[-1] - f[0] ) ( 1 + ------ ) - dr ( 1 + ---- )^2 ( bc_coef - bc_lin * f[0] ) }
    2 r^2                           dr^2                dr



***************************************************************************************/

void FieldObj::nablaSphereR(long i, int ix, long bcvar, 
                      double &rminus, double &rcenter, double &rplus, double &rcoef)
{
  double a, b;
  int bc;

    rcenter = - ( (rminus = dxminus[ix]) + (rplus = dxplus[ix]) );
    rcoef = 0.0;

    if ( bcvar & SURF_XMIN )
      {
      rminus = 0.0;
      if ( bc_deriv[bc = (bcvar >> XSHIFT) & BC_ID_MASK ] == 0)  
        rcoef = dxminus[ix] * bc_coef[bc];
      else 
        if (xcoord[ix] > 0 ) { 
          a = 2 * xcoord[ix] * dxinv[ix];
          b = 0.5 / ( xcoord[ix] * xcoord[ix] );
          rplus = b * (1 + a * a );
          rcoef = - b * (1 - a) * (1 - a) * xgrid[ix]; 
          rcenter = - (rplus + rcoef * bc_lin[bc]);
          rcoef *= bc_coef[bc];
 	  }
        else rcenter = - ( rplus = 6 * dx2inv[ix] );
      }
    else if ( bcvar & SURF_XMAX ) 
      {
      rplus = 0.0;
      if ( bc_deriv[bc = (bcvar >> XSHIFT) & BC_ID_MASK ] == 0)  
         rcoef = dxplus[ix] * bc_coef[bc];
      else 
        {
        a = 2 * xcoord[ix] * dxinv[ix-1];
        b = 0.5 / ( xcoord[ix] * xcoord[ix] );
        rminus = b * ( 1 + a * a );
        rcoef = - b * (1 + a) * (1 + a) * xgrid[ix-1]; 
        rcenter = - (rminus + rcoef * bc_lin[bc]);
        rcoef *= bc_coef[bc];
        }
      }

}

/***************************************************************************************

                        SPHERICAL COORDINATES THETA DIFFERENTIAL
                        ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

                         1       d              df
 Dth^2 f[j,i] == -------------- ---- ( sin(th) ---- ) == dyminus[i] f[i-1] + dycenter[i] f[i] + dyplus[i] f[i+1]
                 r[j]^2 sin(th) d th           d th

 where th == theta[i]
 
        1      sin(th[i+1/2]) (f[i+1] - f[i]) / dth[i] - sin(th[i-1/2]) (f[i] - f[i-1]) / dth[i-1]
 = ----------- ---------------------------------------------------------------------------------
   r^2 sin(th)                               (dth[i] + dth[i-1]) / 2



 At Neumann boundary:  f[j-1] = f[j+1] - 2 dygrid[j] (bc_coef - bc_lin f[j]),

                       dth[i] = dth[i-1] == dth

                 2          / (f[j+1] - f[j]) sin(theta) cos(dth/2)                                        \
 Dth^2 f = --------------- |  ------------------------------------- - sin(th-dth/2) (bc_coef - bc_lin f[j]) |
           r^2 sin(th) dth  \                 dth                                                          /

  Note that the "sin(theta)" cancels in the first term


 ***************************************************************************************/



void FieldObj::nablaTheta(long i, int iy, long bcvar, 
                      double &thminus, double &thcenter, double &thplus, double &thcoef)
{
  int bc;

    thcenter = - ( (thminus = dyminus[iy]) + (thplus = dyplus[iy]) );
    thcoef = 0.0;

    if (bcvar & SURF_YMIN) {
        thminus = 0.0;
        if ( bc_deriv[bc = (bcvar >> YSHIFT) & BC_ID_MASK ] == 0)  
          thcoef = dyminus[iy] * bc_coef[bc];
        else 
          if (ycoord[iy] > 0) { 
            thplus = 2 * cos( 0.5 * ygrid[iy] ) * dy2inv[iy];
            thcoef = - 2 * dyinv[iy] * sin( ycoord[iy] - 0.5 * ygrid[iy]) / sin( ycoord[iy] ) ;
            thcenter = - (thplus + thcoef * bc_lin[bc]);
            thcoef *= bc_coef[bc];
 	    }
	  else    // theta = 0 
            thcenter = - (thplus = 4 * dy2inv[iy]);
    }
    else  if (bcvar & SURF_YMAX) {
        thplus = 0.0;
        if ( bc_deriv[bc = (bcvar >> YSHIFT) & BC_ID_MASK ] == 0)  
          thcoef = dyplus[iy] * bc_coef[bc];
        else 
          if ( fabs( ycoord[iy] - pi) > 1.0e-12 ) { 
            thminus = 2 * cos( 0.5 * ygrid[iy-1] ) * dy2inv[iy-1];
            thcoef = - 2 * dyinv[iy-1] * sin( ycoord[iy] + 0.5 * ygrid[iy-1]) / sin( ycoord[iy] ) ;
            thcenter = - (thminus + thcoef * bc_lin[bc]);
            thcoef *= bc_coef[bc];
 	    }
	  else   // theta =  pi 
            thcenter = - (thminus = 4 * dy2inv[iy-1]);
    }

}

//**********************************************************************************************
/*

 void FieldObj::cleanBoundaries()
    { 
    for (long i = 0; i < size; i++)  
      if ( !(ptype[i] & VARY_MASK) )
        elem[i] = (ptype[i] & DIRICHLET) ? bc_coef[ptype[i] >> XSHIFT] : 0;
    }
*/
//***************************************************************

void FieldObj::print_ptype(long i)
{
long x = ptype[i];
if (x & _OUTSIDE_) fprintf(stdlog,"[OUT] ");
else
  {
  if (x & VARY_MASK) fprintf(stdlog,"[VARY");
  else if (x & DIRICHLET) fprintf(stdlog,"[f=%g", bc_coef[x >> XSHIFT]); 
  if (x & SURF_MASK)
    {
    if (x & SURF_XMIN) fprintf(stdlog," %s XMIN", bc_id[(x >> XSHIFT) & BC_ID_MASK]);
    if (x & SURF_XMAX) fprintf(stdlog," %s XMAX", bc_id[(x >> XSHIFT) & BC_ID_MASK]);
    if (x & SURF_YMIN) fprintf(stdlog," %s YMIN", bc_id[(x >> YSHIFT) & BC_ID_MASK]);
    if (x & SURF_YMAX) fprintf(stdlog," %s YMAX", bc_id[(x >> YSHIFT) & BC_ID_MASK]);
    if (x & SURF_ZMIN) fprintf(stdlog," %s ZMIN", bc_id[(x >> ZSHIFT) & BC_ID_MASK]);
    if (x & SURF_ZMAX) fprintf(stdlog," %s ZMAX", bc_id[(x >> ZSHIFT) & BC_ID_MASK]);
    }
   fprintf(stdlog,"] ");
  }
}


//****************************************************************************
//*                   C L A S S   B U F F E R   A R R A Y
//****************************************************************************

  BufferObj :: BufferObj(TokenString &params, const char *id) : 
  FieldObj(params, id)
    {
      if (total <= 0) {
        char temp[1024];
        sprintf(temp,"%s.%s",id, "total");
        params.errorMessage( params.token2_index(temp, ASSIGN_TOKEN) + 1, 0,
                             "Total buffer concentration must be greater than zero");
      }

    KD = kplus = kminus = -1.0;

    params.get2_param(id, "koff",  &kminus);
    params.get2_param(id, "kminus",&kminus);
    params.get2_param(id, "kon",   &kplus); 
    params.get2_param(id, "kplus", &kplus); 
    params.get2_param(id, "KD",    &KD);  

    if (kplus > 0 && kminus >= 0) KD = kminus / kplus;
    else if (kplus >= 0 && KD >= 0) kminus = kplus * KD;
    else if (kminus >= 0 && KD > 0)  kplus  = kminus / KD;
    else throw makeMessage("Invalid or undefined binding kinetics for buffer \"%s\" (on rate = %g, off rate = %g, KD = %g)",
                  ID, kplus, kminus, KD);

    if ( fabs(KD * kplus - kminus) > 1e-9 ) 
         throw makeMessage("Incompatible properties for buffer \"%s\": on rate = %g, off rate = %g, KD = %g",
                            ID, kplus, kminus, KD);

    if ( !params.token_count(id,"bgr") ) {
      double calcium = params.get_param("Ca.bgr");
      bgr = total / (1.0 + kplus / kminus * calcium);
      if (bgr < 0) params.errorMessage(params.token2_index("Ca.bgr", ASSIGN_TOKEN) + 1, 0,
                                       "Background Ca concentration cannot be negative" );
      if (VERBOSE) fprintf(stdlog,"    Free background [%s] = %g uM\n", ID, bgr);
    }

    if (VERBOSE) 
      fprintf(stdlog,"    Kinetics for buffer %s: KD=%g uM, k-=%g /ms, k+=%g /(ms uM)\n",
	             ID, KD, kminus, kplus);

    if (!params.token_count(id, "import")) *this = bgr;

    for (int j = 0; j < Region->get_surface_num(); j++)
        if (bc_deriv[bccond[j]] == 0)  bc_coef[bccond[j]] = bgr;

    // initialAve = bgr;
    }

//****************************************************************************

BufferArray::BufferArray(TokenString &params, const char *group) {
    
    ID = StrCpy( group );
    buf_num = params.token_count(group);
    array = new BufferObj *[buf_num];

    if ( !buf_num )  return;

    for (int i=0; i<buf_num; i++) {
      long p = params.token_index(ID, i+1) + 1;
      char *id = params.getVarName( p, group );
      for (int j = i-1; j>0; j--) 
        if ( equal( id, array[j]->ID) ) 
           params.errorMessage(p+1, makeMessage("Buffer \"%s\" defined more than once",ID) );
      array[i] = new BufferObj(params, id);
      delete [] id;
    }
}

//****************************************************************************

void BufferArray::allocateCopy(const BufferArray &bufs)
    {
    if (bufs.ID) ID = StrCpy( bufs.ID );
    if ( (buf_num = bufs.buf_num) > 0 )
      {
      array = new BufferObj *[buf_num];
      for (int i = 0; i < buf_num; i++) array[i] = new BufferObj(*bufs.array[i]);
      }
    }

//****************************************************************************

double calcium_gain( FieldObj *Ca, BufferArray *Bufs) {
  double x = 0.0;
  if (Ca) x = Ca->gain();
  if (Bufs) x -= Bufs->gain();
  return x;
}

//****************************************************************************
